//----------------------------------------------------------------------------
// 
//	file: spikelayer3.cc
// 
//	function: make spikelayer more robust to parameter change 
//		- multiplicative dynamic threshold
//		- no self-exc, self-inh
//	author:   Yoonsuck Choe <yschoe@cs.utexas.edu>
//	creation: Fri Nov 27 04:52:03 CST 1998
// 
//	$Revision: 1.2 $ $Date: 2001/11/30 20:24:15 $
//	$Log: spikelayer3.cc,v $
//	Revision 1.2  2001/11/30 20:24:15  yschoe
//	pglissom v1.0 release
//
//	Revision 1.1  1999/12/22 05:29:48  yschoe
//	Initial revision
//
//	Revision 1.1  1999/02/09 07:42:25  yschoe
//	Initial revision
//
//
//----------------------------------------------------------------------------

#include "spikelayer3.h"

//----------------------------------------------------------------------------
//	spikelayer3() 
//	  - Constructor
//----------------------------------------------------------------------------
spikelayer3::spikelayer3(char *name, int cols, int rows, 
		 int exc_rad, int inh_rad,
                 int max_inp_layers, netlayer *feeding,
                 int inp_cols, int inp_rows, int  aff_rad,
                 REAL synapse_decay, REAL inhloop_decay, REAL theta, REAL tau)
	: spikelayer2(name, cols, rows, exc_rad, inh_rad, max_inp_layers,
		 feeding, inp_cols, inp_rows, aff_rad, synapse_decay, 
		 inhloop_decay, theta, tau) {

  // use normalized LUTs (i.e., the leaky sums are divided by the number of
  //	bits)
  delete [] synapse_lut;
  delete [] inhloop_lut;

  synapse_lut  = init_leaky_table(SPIKE_QLEN,synapse_decay);
  inhloop_lut  = init_leaky_table(SPIKE_QLEN,inhloop_decay);

  cout << "SPIKE_QLEN = " << SPIKE_QLEN << endl;

}

//----------------------------------------------------------------------------
//	activate() 
//	  - activate spike layer
//----------------------------------------------------------------------------
void spikelayer3::activate(int calc_aff) {

  // check if recursive lateral connections are killed
  static int rec_lat_killed = 0;
  if (!rec_lat_killed) {
	kill_recursive_lat();
  }

  // 0. calc and push spikes
  push_spike();

  // 1. use netlayer::calc_affresp() to calculate net_input.
  //    todo: If we assume that input spikes reach a stable firing before
  //    feeding into the cortex, we can avoid calculating affresp 
  //    everytime during settling.
  calc_affresp();       // net_aff_input <-- wgt * (feeding->output); 

  // 2. use netlayer::calc_latresp() to calculate lateral contrib
  calc_latresp();       // net_exc/inh_input <- wgt*output (from previous step)
  calc_contrib_sum(0);  // net_input <-- net_aff_input + net_exc/inh_input

  // SPIKELAYER2 : pass net_input through sigmoid : output <- sigmoid(net_input)
  apply_actfunc();

  // 3. calculate spike
  calc_spike();         // output <-- shifted, decayed sum of spikes

}

//----------------------------------------------------------------------------
//      calc_spike() 
//        - generate spike and push spike, put LUT value, etc.
//----------------------------------------------------------------------------
void spikelayer3::calc_spike() {

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

        // 0. just for displaying -- never used in calculation 
        lfp[i][j] = output[i][j];  // this was calculated in activate().

        // 1. generate spike
        // *. spikeact_q is identical to inhloop_q, so there's
        //    spikeact_q is used instead of inhloop_q when looking up the
        //    inhloop_lut.
        spike[i][j] = ( output[i][j]
                > theta * (1.0+tau*inhloop_lut[(unsigned)spikeact_q[i][j] & SPIKE_QMASK]) + intrinsic_noise*drand48()
            ) ? 1 : 0;
        /* DEBUG
        if (i==2 && j==2) {
                disp_binary(spikeact_q[i][j] & SPIKE_HISTQMASK);
                cout << "  :  " 
                     << ((unsigned)spikeact_q[i][j] & SPIKE_HISTQMASK) ;
        }*/

     }
  }

}

//----------------------------------------------------------------------------
//	kill_recursive_lat() 
//	  - kill recursive lateral connections
//----------------------------------------------------------------------------
void spikelayer3::kill_recursive_lat(){

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

        // 1.1 excitatory
        int lowx = lat_exc_range[i][j].lowx;
        int lowy = lat_exc_range[i][j].lowy;
	wgt[i][j].lat_exc_w[i-lowx][j-lowy] = -DEATH_FLAG;

        // 1.2 inhibitory
        lowx = lat_inh_range[i][j].lowx;
        lowy = lat_inh_range[i][j].lowy;
        wgt[i][j].lat_inh_w[i-lowx][j-lowy] = -DEATH_FLAG;

     }
  }

}
//----------------------------------------------------------------------------
//      print_stat() 
//        - print status of (col,row) neuron.
//----------------------------------------------------------------------------
void spikelayer3::print_stat(int cur_inp, int col, int row) {

  cout  << net_aff_input[col][row] << " "
        << (lat_exc_contri*net_exc_input[col][row]) << " "
        << (lat_inh_contri*net_inh_input[col][row]) << " "
        << net_input[col][row] << " "
        << (synapse_lut[(unsigned)spikeact_q[col][row]&SPIKE_QMASK ]) << " "
        << theta << " "
        << (theta + net_input[col][row]-theta*(1.0+tau*inhloop_lut[(unsigned)spikeact_q[col][row]&SPIKE_QMASK]))<<endl;

}

